#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <glm/geometric.hpp>
#include <iostream>
#include <map>
#include <memory>
#include <tuple>
#include <vector>

#include "BulletCollision/CollisionDispatch/btCollisionObject.h"
#include "BulletCollision/CollisionShapes/btBoxShape.h"
#include "LinearMath/btVector3.h"
#include "constants.h"
#include "engine.h"
#include "gameobject.h"
#include "input.h"
#include "internal/physics.h"
#include "kekengine.h"
#include "light.h"
#include "mesh.h"
#include "player.h"
#include "scene.h"
#include "texture.h"
#include "ui.h"
#include "uielements.h"
#include "utils.h"

using namespace kek;

#define CHUNK_SIZE 64
#define CHUNK_RADIUS 2
#define WALLS_PER_CHUNK 20
#define WALL_HEIGHT 10
#define LIGHT_SPACING 16

static ButtonElement *buttonPlay;

static SpotLight *flashlight;

struct Chunk {
	int x, z;
	GameObject *floor;
	GameObject *ceiling;
	std::vector<GameObject *> walls;
	std::vector<PointLight *> lights;
};

std::map<long int, Chunk> loadedChunks;
Mesh *floorMesh;
TextElement *debugText;
std::shared_ptr<Texture> red, gray, white, wallpaper;

void gameLoop(kek::PeriodicEvent event) {
	Player *player = Engine::getPlayer();
	Camera *camera = Engine::getActiveCamera();
	glm::vec3 playerPos = player->getPosition();
	Scene *scene = Engine::getActiveScene();

	glm::vec3 playerRight = glm::cross(camera->direction, glm::vec3(0, 1, 0));
	flashlight->moveTo(playerPos + glm::normalize(playerRight) * 0.3f);
	flashlight->lookAt(camera->direction);

	int chunkX = (int) playerPos.x / CHUNK_SIZE;
	int chunkZ = (int) playerPos.z / CHUNK_SIZE;
	debugText->setText("CX: " + std::to_string(chunkX) + ", CZ: " + std::to_string(chunkZ));

	for(auto it = loadedChunks.cbegin(); it != loadedChunks.cend();) {
		const Chunk *ch = &it->second;
		if(std::abs(ch->x - chunkX) > CHUNK_RADIUS || abs(ch->z - chunkZ) > CHUNK_RADIUS) {
			scene->removeObject(ch->floor);
			scene->removeObject(ch->ceiling);

			for(auto w : ch->walls) {
				scene->removeObject(w);
			}

			for(auto l : ch->lights) {
				scene->lights->remove(l);
			}

			it = loadedChunks.erase(it);
			continue;
		}

		it = std::next(it);
	}

	for(int x = chunkX - CHUNK_RADIUS; x <= chunkX + CHUNK_RADIUS; x++) {
		for(int z = chunkZ - CHUNK_RADIUS; z <= chunkZ + CHUNK_RADIUS; z++) {
			long int chunkID = ((long int) x & 0xFFFF) << 16 | ((long int) z & 0xFFFF);
			auto chunk = loadedChunks.find(chunkID);
			if(chunk == loadedChunks.end()) {
				GameObject *floor = new GameObject();
				floor->addMesh(genCubeMesh(CHUNK_SIZE, 2, CHUNK_SIZE, gray, gray, gray));
				floor->addPhysics(new btBoxShape(btVector3(CHUNK_SIZE / 2, 1, CHUNK_SIZE / 2)), 0, btCollisionObject::CF_STATIC_OBJECT);
				floor->moveTo(glm::vec3(x * CHUNK_SIZE, 0, z * CHUNK_SIZE));
				scene->addObject(floor);

				GameObject *ceiling = new GameObject();
				ceiling->addMesh(genCubeMesh(CHUNK_SIZE, 1, CHUNK_SIZE, gray, gray, gray));
				// ceiling->addPhysics(new btBoxShape(btVector3(CHUNK_SIZE / 2, 1, CHUNK_SIZE / 2)), 0, btCollisionObject::CF_STATIC_OBJECT);
				ceiling->moveTo(glm::vec3(x * CHUNK_SIZE, WALL_HEIGHT, z * CHUNK_SIZE));
				scene->addObject(ceiling);

				std::vector<GameObject *> walls;
				std::srand(chunkID);
				for(unsigned int i = 0; i < WALLS_PER_CHUNK; i++) {
					double wallLength = (double) std::rand() / RAND_MAX * 20 + 10;
					bool rotated = std::rand() < RAND_MAX / 2;
					GameObject *wall = new GameObject();
					wall->addMesh(genCubeMesh(wallLength, WALL_HEIGHT, 1, wallpaper, wallpaper, wallpaper));
					wall->addPhysics(new btBoxShape(btVector3(wallLength / 2, WALL_HEIGHT / 2, 0.5)), 0, btCollisionObject::CF_STATIC_OBJECT);
					wall->moveTo(glm::vec3(x * CHUNK_SIZE + ((double) rand() / RAND_MAX * CHUNK_SIZE), WALL_HEIGHT / 2, z * CHUNK_SIZE + ((double) rand() / RAND_MAX * CHUNK_SIZE)));

					if(rotated) {
						wall->rotate(glm::radians(90.0f), glm::vec3(0, 1, 0));
					}

					scene->addObject(wall);
					walls.push_back(wall);
				}

				std::vector<PointLight *> lights;
				for(int lx = 0; lx < CHUNK_SIZE / LIGHT_SPACING; lx++) {
					for(int ly = 0; ly < CHUNK_SIZE / LIGHT_SPACING; ly++) {
						// SpotLight *light = new SpotLight(glm::vec3(1), glm::vec3(0, -1, 0), 1, 1, 1, 0.5, 0.6);
						glm::vec3 lightPos(x * CHUNK_SIZE + lx * LIGHT_SPACING, WALL_HEIGHT - 0.5, z * CHUNK_SIZE + ly * LIGHT_SPACING);

						PointLight *light = new PointLight(glm::vec3(1), 1, 1, 1);
						light->moveTo(glm::vec3(x * CHUNK_SIZE + lx * LIGHT_SPACING, WALL_HEIGHT - 1, z * CHUNK_SIZE + ly * LIGHT_SPACING));
						scene->lights->add(light);
						lights.push_back(light);

						GameObject *lamp = new GameObject();
						lamp->addMesh(genCubeMesh(2, 1, 2, white, white, white));
						lamp->moveTo(lightPos);
						walls.push_back(lamp);
						scene->addObject(lamp);
					}
				}

				Chunk newChunk{x, z, floor, ceiling, walls, lights};
				loadedChunks.emplace(chunkID, newChunk);
			}
		}
	}
}

void startGame() {
	buttonPlay->visible = false;

	Scene *scene = new Scene();

	wallpaper = Texture::load("image/wallpaper.png");
	white = Texture::generateColor(glm::vec3(1));
	red = Texture::generateColor(glm::vec3(1, 0, 0));
	gray = Texture::generateColor(glm::vec3(0.2, 0.2, 0.2));

	flashlight = new SpotLight(glm::vec3(1), glm::vec3(0), 2, 3, 5, 0.5, 0.6);
	flashlight->moveTo(Engine::getPlayer()->getPosition() + glm::vec3(0, KEK_PLAYER_EYE_OFFSET, 0));

	Input::addPeriodicCallback(PeriodicCallback(gameLoop));

	std::shared_ptr<Texture> cubeTexture = Texture::generateColor(glm::vec3(1, 0, 0));
	GameObject *floorObject = new GameObject();
	floorObject->addPhysics(new btBoxShape(btVector3(CHUNK_SIZE / 2, 0.5, CHUNK_SIZE / 2)), 0, btCollisionObject::CF_STATIC_OBJECT);
	floorObject->addMesh(genCubeMesh(CHUNK_SIZE, 1, CHUNK_SIZE, cubeTexture, cubeTexture, white));
	scene->addObject(floorObject);

	for(unsigned int i = 0; i < 10; i++) {
		GameObject *cube = new GameObject();
		cube->addPhysics(new btBoxShape(btVector3(0.5, 0.5, 0.5)), 1);
		cube->addMesh(genCubeMesh(1, 1, 1, wallpaper, wallpaper, white));
		cube->moveTo(glm::vec3(5, i * 1.1f, 0));
		scene->addObject(cube);
	}

	scene->lights->add(flashlight);
	// scene->lights->add(new DirectionalLight(glm::vec3(0.005), glm::vec3(-0.1, -1, 0.1)));

	debugText = new TextElement(uiPx(10), uiPx(100));
	debugText->setText("Hello!");
	UI::addElement(debugText);

	Engine::setActiveScene(scene);

	Engine::getPlayer()->moveTo(glm::vec3(0, KEK_PLAYER_HEIGHT - KEK_PLAYER_RADIUS, 0));
	Engine::getActiveCamera()->lookAt(glm::vec3(0, 0, 1));

	Input::setCursorMode(GLFWCursorMode::CAPTURE);
}

int main() {
	if(Engine::init() != KEK_SUCCESS) return 1;

	buttonPlay = new ButtonElement(uiSw(0.5), uiSh(0.5), uiPx(500), uiPx(75));
	buttonPlay->color = Colors::WHITE;
	buttonPlay->hoverColor = Colors::GRAY;
	buttonPlay->origin = Origin::CENTER;
	buttonPlay->text->setText("Play");
	buttonPlay->text->sizePixels = 50;
	buttonPlay->text->color = Colors::BLACK;
	buttonPlay->onClick = Callable(startGame);

	UI::addElement(buttonPlay);
	Input::setCursorMode(GLFWCursorMode::FREE);

	if(Engine::start() != KEK_SUCCESS) return 1;

	Engine::exit();
}
